/**

    Copyright 2007, 2008 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
**/
package org.spagic.ui.actions;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.emf.common.util.EMap;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EAnnotation;
import org.eclipse.emf.ecore.EcoreFactory;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.dialogs.ErrorDialog;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.stp.bpmn.Activity;
import org.eclipse.stp.bpmn.BpmnDiagram;
import org.eclipse.stp.bpmn.Pool;
import org.eclipse.stp.bpmn.SubProcess;
import org.eclipse.stp.bpmn.Vertex;
import org.eclipse.stp.im.constants.ImConstants;
import org.spagic.ui.SpagicUIActivator;
import org.spagic.ui.dialog.ProcessPropertiesDialog;
import org.eclipse.stp.im.util.ImLogger;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IActionDelegate;
import org.eclipse.ui.IObjectActionDelegate;
import org.eclipse.ui.IWorkbenchPart;


public class ConfigurePropertiesAction implements IObjectActionDelegate {

	private IFile bpmnFile = null;
	
	public static final String PARAMETER_START = "${".intern();
	public static final String PARAMETER_END = "}".intern();
	
	/**
	 * Constructor for ConfigurePropertiesAction.
	 */
	public ConfigurePropertiesAction() {
		super();
	}

	/**
	 * @see IObjectActionDelegate#setActivePart(IAction, IWorkbenchPart)
	 */
	public void setActivePart(IAction action, IWorkbenchPart targetPart) {
	}

	/**
	 * @see IActionDelegate#run(IAction)
	 */
	public void run(IAction action) {
		final Shell shell = new Shell();

		try {
			configureParameters(shell);
		} catch (Exception e) {
			reportError(shell, e);
		}

	}
	
    /**
  	 * Creates an EAnnotation with a source parameter and
  	 * a details map.
  	 * @param source
  	 * @param details
  	 * @return
  	 */
  	protected EAnnotation createAnnotation(String source, Map<String, String> details) {
  		EAnnotation ea = EcoreFactory.eINSTANCE.createEAnnotation();
  		ea.setSource(source);
  		ea.getDetails().putAll(details);
  		return ea;
  	}  	

  	public void configureParameters(Shell shell) throws Exception {
		// Read the bpmn emf model
		ResourceSet resourceSetBpmn = new ResourceSetImpl();
		URI uri = URI.createPlatformResourceURI(bpmnFile.getFullPath().toString(), true);
		
		Resource resourceBpmn = resourceSetBpmn.getResource(uri, true);	
		BpmnDiagram bpmnDiagram = (BpmnDiagram)resourceBpmn.getContents().get(0);
		
		// Retrieve parameters used in BPMN diagram nodes
		Properties bpmnParameters = retrieveBPMNParameters(bpmnDiagram);

		
		// Retrieve parameters values already configured in BPMN diagram annotation and
		// merge with the previous list
		updateBPMNParameters(bpmnDiagram, bpmnParameters);

		
		ProcessPropertiesDialog aListPropertiesDialog = new ProcessPropertiesDialog(shell, bpmnParameters);
		int result = aListPropertiesDialog.open();
		
		if (result == ProcessPropertiesDialog.OK) {			
			
			// Save all parameters within the pools
			// Actually we save these parameters on all pools: they should be saved in the pool where
			// they are used
			for (Pool aPool : bpmnDiagram.getPools()) {

	    	 	EAnnotation poolParametersAnnotation = aPool.getEAnnotation(ImConstants.PARAMETERS_ANNOTATION);
	        	   
	    	 	if (poolParametersAnnotation == null){
	    	 		// Annotation not existing: create it and fill it
	    	 		Map<String, String> details = new HashMap<String, String>();
	    	 		poolParametersAnnotation = createAnnotation(ImConstants.PARAMETERS_ANNOTATION, details);
	    	 		
	    	 		for (Object key : bpmnParameters.keySet()) {
	    	 			String propertyValue = bpmnParameters.getProperty((String)key);
		    	 		details.put((String)key, propertyValue);
	    	 		}
	    	 		
	    	 		aPool.getEAnnotations().add(poolParametersAnnotation);
	    	 	}else{
	    	 		// Annotation already existing: clean it and refill it again
	    	 		EMap<String, String> details = poolParametersAnnotation.getDetails();
	    	 		details.clear();
	    	 		for (Object key : bpmnParameters.keySet()) {
	    	 			String propertyValue = bpmnParameters.getProperty((String)key);
	    	 			details.put((String)key, propertyValue);
	    	 		}
	    	 	}	   
			}
			
			// Save BPMN diagram
			resourceBpmn.save(Collections.EMPTY_MAP);
		}
	}
	
	
	/**
	 * Retrieves the parameters from the nodes of the BPMN diagram.
	 * The parameters are recognized by the syntax ${parameter name}.
	 * Scan all the properties of all nodes to retrieve the parameters.
	 * @param bpmnDiagram Diagram containing the annotations with parameters values
	 * @return Map of <keys, values>
	 */
	private Properties retrieveBPMNParameters(BpmnDiagram bpmnDiagram) {
		Properties bpmnParameters = new Properties();
		
		for (Pool aPool : bpmnDiagram.getPools()) {
			for (Vertex v : aPool.getVertices()) {
				if (v instanceof Activity) {
					Activity a = (Activity)v;
					retrieveParametersForActivity(a, bpmnParameters);
					
					if (a.isLooping() || a.isSetLooping()){
						System.out.println(" Is Looping "+ a);
						 if (a instanceof SubProcess) {
							// Generate step for each loop node
							for (Vertex vertexSubProcess : ((SubProcess)a).getVertices()){
								retrieveParametersForActivity((Activity)vertexSubProcess, bpmnParameters);
							}
						}
					}
				} else {
					ImLogger.warning(SpagicUIActivator.PLUGIN_ID," Not Recognized v ["+v.getID()+"] as Activity");
				}
			}
		}
		
		return bpmnParameters;
	}
	
	/**
	 * Merge the parameters used in the BPMN diagram, with the parameters already configured
	 * by the user for that diagram.
	 * @param bpmnDiagram BPMN diagram containing the parameters already configured by the user
	 * @param bpmnParameters Parameters used in the nodes of the diagram
	 */
	private void updateBPMNParameters(BpmnDiagram bpmnDiagram, Properties bpmnParameters) {
		
		for (Pool aPool : bpmnDiagram.getPools()) {
    	 	EAnnotation poolParametersAnnotation = aPool.getEAnnotation(ImConstants.PARAMETERS_ANNOTATION);
        	   
    	 	if (poolParametersAnnotation != null) {
    	 		// Annotation existing: read all contained parameters
    	 		EMap<String, String> details = poolParametersAnnotation.getDetails();
    	 		
    	 		for (String key : details.keySet()) {
    	 			String propertyValue = details.get(key);
    	 			
    	 			// If the parameter already exists in bpmnParameters structure
    	 			// this means that it has been used in a node, so we should consider
    	 			// its value.
    	 			// Otherwise we ignore it, because it's a parameter not anymore used
//    	 			if (bpmnParameters.containsKey(key)) {
    	 				bpmnParameters.setProperty(key, propertyValue);
//    	 			}
    	 		}
    	 	}	   
		}
	}
	
	public void retrieveParametersForActivity(Activity a, Properties parameters) {

		EAnnotation imAnnotation = a.getEAnnotation(ImConstants.IM_ANNOTATION);

		if (imAnnotation != null) {

			for (String propertyKey : imAnnotation.getDetails().keySet()) {

				// Ignore basic properties
				if (propertyKey.equalsIgnoreCase(ImConstants.IM_SERVICE_NAME) ||
					propertyKey.equalsIgnoreCase(ImConstants.IM_SERVICE_BINDING_NAME) ||
					propertyKey.equalsIgnoreCase(ImConstants.IM_POOL_RUNTIME_ID)) {
					continue;
				}
				
				String propertyValue = imAnnotation.getDetails().get(propertyKey);
				if (propertyValue != null) {
					int startIndex = propertyValue.indexOf(PARAMETER_START);
					if ((startIndex != -1) && (propertyValue.indexOf(PARAMETER_END, startIndex) != -1)) {
						
						// Split the string only if we are sure that there is at least a parameter
						String[] splitted = propertyValue.split(PARAMETER_END);
						for (String part : splitted) {
							if (part.indexOf(PARAMETER_START) != -1) {
								// Parameter found, no value yet available
								String parameterName = part.substring(part.indexOf(PARAMETER_START) + PARAMETER_START.length());
								parameters.setProperty(parameterName, "");
							}
						}
					}
				}
			}

		} else {
			ImLogger.warning(SpagicUIActivator.PLUGIN_ID, "Activity ["
					+ a.getName() + "] does not have annotation");
		}
	}	
	
	
	public final void selectionChanged(final IAction action,
			final ISelection selection) {
		
		setBpmnFile((IFile) ((StructuredSelection) selection)
				.getFirstElement());
	}

	public IFile getBpmnFile() {
		return bpmnFile;
	}

	public void setBpmnFile(IFile bpmnFile) {
		this.bpmnFile = bpmnFile;
	}
	
	/**
    * Displays an error that occured during the project creation.
    * @param x details on the error
    */
   private void reportError(Shell shell, Throwable x)
   {
      ErrorDialog.openError(shell,
                            "Error",
                            "Error Occurred",
                            makeStatus(x));
   }
   
   
   public static IStatus makeStatus(Throwable x)
   {
	   		
	   return new Status(IStatus.ERROR,
			SpagicUIActivator.PLUGIN_ID,
            IStatus.ERROR,
            x.getMessage(), null);

   }
}
	
	
