package org.spagic.utils;

import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.eclipse.stp.im.BasicProperty;
import org.eclipse.stp.im.ConfigurableElement;
import org.eclipse.stp.im.ImFactory;
import org.eclipse.stp.im.Process;
import org.eclipse.stp.im.Property;
import org.eclipse.stp.im.ServiceBinding;
import org.eclipse.stp.im.Step;
import org.eclipse.stp.im.StpIntermediateModel;
import org.eclipse.stp.im.Transition;
import org.eclipse.stp.im.constants.ImConstants;


public class SpagicJBIImPreprocessor extends AbstractImPreprocessor {

	private static final String IS_STEP_INTERMEDIATE_EVENT = "im.isStepIntermediateEvent";
	private static final String JBI_TECHNOLOGY = "JBI";
	private static final String AGGREGATOR_ENDPOINT = "aggregatorEndpoint";
	
	private static final String REPLY_STEP = "JBI-Reply".intern();
	private static final String HIBERNATE_TRANSACTED_REPLY_STEP = "JBI-HibernateTransactedReply".intern();
	
	private static final String CLASSIC_SYNCHRONIZER = "JBI-Sync".intern();
	private static final String HIBERNATE_TRANSACTED_SYNCHRONIZER = "JBI-TransactedSync".intern();
	
	private static final String WEB_SERVICE_PIPELINE = "JBI-WebServicePipeline".intern();
	private static final String TCP_IP_PIPELINE = "JBI-TCPIP-Pipeline".intern();
	private static final String CXF_PIPELINE = "JBI-CXF-Pipeline".intern();
	
	private static final String HTTP_IO_BC = "JBI-HTTPInputOutputBindingComponent".intern();
	private static final String TCP_IO_BC = "JBI-TcpIp-InputOutputBindingComponent".intern();
	private static final String CXF_IO_BC = "JBI-CXF-InputOutputBindingComponent".intern();
	
	private String runtimeId = null;

	
	public SpagicJBIImPreprocessor(String runtimeId){
		super();
		this.runtimeId = runtimeId;
	}
	
	public void process(StpIntermediateModel im) {
		for (Process process : im.getProcessCollection().getProcesses()) {
			if (!SpagicUtils.isProcessDeployableForRuntime(process, getRuntimeId()))
				continue;
			handleImProcess(im, process);
		}
	}
	
	public void handleImProcess(StpIntermediateModel im, Process process) {
	
		// Create the technology property
		createTechnologyProperty(process, JBI_TECHNOLOGY);
		
		// Handle The Synchronizer Pattern
		handleSynchronizer(im, process);
		
		// Handle Pipeline
		handlePipeline(im, process);
		
		// Handle Wait Event
		handleWaitEvent(im, process);
		
		// Handle the timer starting for Aggregators 
		handleTimerAggregator(im, process);
			
	}
		
	public void createServiceBindingForRuntime(StpIntermediateModel im, String serviceBindingName) {
		
		ServiceBinding serviceBinding = factory.createServiceBinding();
		serviceBinding.setName(serviceBindingName);
		BasicProperty runtimeIdProperty = factory.createBasicProperty();
		runtimeIdProperty.setKey(ImConstants.IM_STEP_BINDING_RUNTIME_ID);
		runtimeIdProperty.setValue(getRuntimeId());
		serviceBinding.getProperties().put(runtimeIdProperty.getKey(),
				runtimeIdProperty);
		im.getServicebindings().add(serviceBinding);
		
	}

	public ServiceBinding getRightStepBinding(Step s) {
	Property runtimPropertyForServiceBinding = null;
	for (ServiceBinding sb : s.getStepbindings()){
		runtimPropertyForServiceBinding = sb.getProperties().get(ImConstants.IM_STEP_BINDING_RUNTIME_ID);
		if (runtimPropertyForServiceBinding != null){
			if (runtimPropertyForServiceBinding.getValue().equals(getRuntimeId())){
				return sb;
			}
		}else{
			SpagicLogger.warning("ServiceBinding["+sb.getName()+"] -> for Step["+s.getName()+"] does not have a property called "+ ImConstants.IM_POOL_RUNTIME_ID);
		}
	}
	SpagicLogger.warning("Not found for Step["+s.getName()+"] a binding for runtime "+ ImConstants.IM_POOL_RUNTIME_ID);
	return null;
	}

	public String getRuntimeId() {
		return runtimeId;
	}

	public void setRuntimeId(String runtimeId) {
		this.runtimeId = runtimeId;
	}
	
	public String getBasicPropertyValue( ConfigurableElement ce, String propertyName ){
		Property p = ce.getProperties().get(propertyName);
		if (p != null){
			 if (p instanceof BasicProperty) {
				return ((BasicProperty)p).getValue();	
			}else{
				return p.toString();
			}
		}else{
			return null;
		}
	}
	public void handleWaitEvent(StpIntermediateModel im, Process aProcess ){
		List<Step> stepToAddToProcess = new ArrayList<Step>();
		List<Step> replyStepsToRemove = new ArrayList<Step>();
		
		for (Step aStep : aProcess.getSteps()) {
			if ( isServiceBinding(aStep, "JBI-HumanTaskEvent") || isServiceBinding(aStep,"JBI-WaitExternalEvent")) {
				
				if (aStep.getSourceTransitions().size() > 1){
					throw new RuntimeException(" A Wait or HumanTaskEvent Binding could have only one destination ");
				}
				
				Transition destinationAfterWaitTransition = getFirstOutcomingTransition(aStep);
				if (destinationAfterWaitTransition == null)
					throw new RuntimeException(" A Wait or HumanTaskEvent Service Must have one destination ");
				
				Step destinationAfterWaitStep = destinationAfterWaitTransition.getTarget();
				
				String signalStepName = aStep.getName() + "Signal";
				
				// Create Signal Step
				Step signalStep = createStep(im, signalStepName, "JBI-SignalEvent");
				stepToAddToProcess.add(signalStep);
				
				String destinationURI = SpagicUtils.getStepNameInProcess(aProcess, destinationAfterWaitStep);
				String signalURI = SpagicUtils.getStepNameInProcess(aProcess, signalStep);
				
				setPropertyOnStep(aStep, "destinationURI", destinationURI);
				setPropertyOnStep(aStep, "signalEndpointURI", signalURI);
				
				setPropertyOnStep(signalStep, "destinationURI", destinationURI);
				
				
				destinationAfterWaitTransition.setSource(signalStep);
				
				createTransitionAndAddToProcess(aProcess, aStep, signalStep);
				
				Transition incomingFromExternalBindingTransition = findTheTransitionComingFromExternalEventHandler(aStep);				
				if ( incomingFromExternalBindingTransition == null ){
					 
					 if (isServiceBinding(aStep,"JBI-WaitExternalEvent")) {
						 SpagicLogger.error(" When using a Wait External Event In The Process must be present a service binding belonging to External Event Service");
					 }else if (isServiceBinding(aStep,"JBI-HumanTaskEvent")) {
						 setPropertyOnStep(aStep, "EventType", "GenericEvent");
					 }
					 
				} else {
					
					if (isServiceBinding(aStep,"JBI-HumanTaskEvent")) {
						 SpagicLogger.error(" When using a JBI-HumanTaskEvent no External Event Service is Required");
					}else if (isServiceBinding(aStep,"JBI-WaitExternalEvent")) {
						 
						 Step incomingExternalServiceStep = incomingFromExternalBindingTransition.getSource();
						 String handledEventType = getEventTypeHandled(incomingExternalServiceStep);
						 setPropertyOnStep(aStep, "EventType",handledEventType );
						 setSignalEndpointURIOnStep(aProcess, incomingExternalServiceStep, signalStep);
						 
						 // Create The Matching Engine Step
						 String eventMatcherServiceBindingName = "JBI-EventMatcher";
						 Step eventMatcherStep = createStep(im, aStep.getName() + "EventMatcher", eventMatcherServiceBindingName);
						 setPropertyOnStep(eventMatcherStep, "signalEndpointURI", signalURI);
						 stepToAddToProcess.add(eventMatcherStep);
						 
						 
						 // Copy on Event Matcher the Matching rules
						 String numberOfRulesOnEventAttributeString = getBasicPropertyValue(incomingFromExternalBindingTransition, "NumberOfRulesOnEventAttribute");
						 String numberOfRulesOnEventPayloadString = getBasicPropertyValue(incomingFromExternalBindingTransition, "NumberOfRulesOnEventPayload");
						 
						 
						 
						 if (numberOfRulesOnEventAttributeString != null){
							 int numberOfRulesOnEventAttribute = Integer.valueOf(numberOfRulesOnEventAttributeString);
							 setPropertyOnStep(eventMatcherStep, "NumberOfRulesOnEventAttribute", numberOfRulesOnEventAttributeString);
							 
							 String pName,  pValue = null;
							 for ( int i = 0; i < numberOfRulesOnEventAttribute; i++){
								 pName = MessageFormat.format("RulesOnEventAttibute[{0}].WaitInstanceAttributeName",i);
								 pValue =  getBasicPropertyValue(incomingFromExternalBindingTransition,pName);
								 setPropertyOnStep(eventMatcherStep, pName, pValue);
								 pName = MessageFormat.format("RulesOnEventAttibute[{0}].EventAttributeName",i);
								 pValue =  getBasicPropertyValue(incomingFromExternalBindingTransition,pName);
								 setPropertyOnStep(eventMatcherStep, pName, pValue);
									
							 }
						 }
						 
						 if (numberOfRulesOnEventPayloadString != null){
							 int numberOfRulesOnEventPayload = Integer.valueOf(numberOfRulesOnEventPayloadString);
							 setPropertyOnStep(eventMatcherStep, "NumberOfRulesOnEventPayload", numberOfRulesOnEventPayloadString);
							 
							 String pName,  pValue = null;
							 for ( int i = 0; i < numberOfRulesOnEventPayload; i++){
								 pName = MessageFormat.format("RulesOnEventPayload[{0}].WaitInstanceAttributeName",i);
								 pValue =  getBasicPropertyValue(incomingFromExternalBindingTransition,pName);
								 setPropertyOnStep(eventMatcherStep, pName, pValue);
								 pName = MessageFormat.format("RulesOnEventPayload[{0}].XPathExpression",i);
								 pValue =  getBasicPropertyValue(incomingFromExternalBindingTransition,pName);
								 setPropertyOnStep(eventMatcherStep, pName, pValue);
							 }
						 }
					     
					     // Relocating Transitions
						 if (handledEventType.equalsIgnoreCase("WebserviceEvent")){

							 String eventWrapperStepName = "JBI-WrapToEvent";
							 Step eventWrapperStep = createStep(im, incomingExternalServiceStep.getName() + "EventWrapper", eventWrapperStepName);
							 setPropertyOnStep(eventWrapperStep, "EventType",handledEventType );
							 String eventMatcherURI = SpagicUtils.getStepNameInProcess(aProcess, eventMatcherStep);
							 setPropertyOnStep(eventWrapperStep, "eventMatcherURI", eventMatcherURI);
							 setPropertyOnStep(eventWrapperStep, IS_STEP_INTERMEDIATE_EVENT, "true");
							 stepToAddToProcess.add(eventWrapperStep);
							 incomingFromExternalBindingTransition.setTarget(eventWrapperStep);
							 createTransitionAndAddToProcess(aProcess, eventWrapperStep, eventMatcherStep);
							 createTransitionAndAddToProcess(aProcess, eventMatcherStep, signalStep);
							 
						 }else{
							 incomingFromExternalBindingTransition.setTarget(eventMatcherStep);
							 createTransitionAndAddToProcess(aProcess, eventMatcherStep, signalStep);
						 }
					}
					
				} //if ( incomingFromExternalBindingComponent == null )
			} // if ( isServiceBinding(aStep, "JBI-HumanTaskEvent") || isServiceBinding(aStep,"JBI-WaitExternalEvent"))
		}
		aProcess.getSteps().removeAll(replyStepsToRemove);		
		aProcess.getSteps().addAll(stepToAddToProcess);
	} //public void handleWaitEvent(StpIntermediateModel im, Process aProcess )

	public void handlePipeline(StpIntermediateModel im, Process aProcess ){

		List<Step> stepToAddToProcess = new ArrayList<Step>();
		List<Step> replyStepsToRemove = new ArrayList<Step>();
		
		Step transformerStep = null;
		for (Step pipelineStep : aProcess.getSteps()) {
			if (isServiceBinding(pipelineStep, WEB_SERVICE_PIPELINE) || isServiceBinding(pipelineStep,TCP_IP_PIPELINE) || isServiceBinding(pipelineStep,CXF_PIPELINE)) {
			
				String serviceBindingName = null;
				if (isServiceBinding(pipelineStep, WEB_SERVICE_PIPELINE)){
					serviceBindingName = HTTP_IO_BC;
				}
				
				if (isServiceBinding(pipelineStep, TCP_IP_PIPELINE)){
					serviceBindingName = TCP_IO_BC;
				}
				
				if (isServiceBinding(pipelineStep, CXF_PIPELINE)){
					serviceBindingName=CXF_IO_BC;
				}
				
				transformerStep = createStep(im, pipelineStep.getName()
						+ SpagicUtilsConstants.JBI_TRANSFORMER_SUFFIX, serviceBindingName);
				stepToAddToProcess.add(transformerStep);
				
				copyAllProperties(pipelineStep, transformerStep);
				

				setPropertyOnStep(pipelineStep, "transformerStepName", transformerStep.getName());
				setPropertyOnStep(transformerStep, "IsTransformerPipeline", "true");
				
				createTransitionAndAddToProcess(aProcess, pipelineStep, transformerStep);				
				createTransitionAndAddToProcess(aProcess, transformerStep, pipelineStep);

			}
		}// end
		aProcess.getSteps().removeAll(replyStepsToRemove);		
		aProcess.getSteps().addAll(stepToAddToProcess);
	}
	
	
	public boolean isCandidateForSynchronizer(Step step){
		
		ServiceBinding sb  = getRightStepBinding(step);
		return (sb != null) && 
			( isServiceBinding(sb,HTTP_IO_BC) || isServiceBinding(sb, TCP_IO_BC) || isServiceBinding(sb, CXF_IO_BC));
	}
	
	public Step createSynchronizerStepForStartStep(StpIntermediateModel im, Step sourceOfSynchronizer, String synchronizerType){
		return createStep(im, sourceOfSynchronizer.getName()
				+ SpagicUtilsConstants.JBI_SYNC_SUFFIX, synchronizerType);
	}
	
	public void handleSynchronizer(StpIntermediateModel im,  Process process) {
		boolean syncAdded = false;
		
		List<Step> stepToAddToProcess = new ArrayList<Step>();
		List<Step> replyStepsToRemove = new ArrayList<Step>();
		ImFactory stpModelfactory = ImFactory.eINSTANCE;
		
	
		Transition sourceTransition = null;
		Step destinationStep = null;
		Step synchronizerStep = null;
		for (Step aStep : process.getSteps()) {
			
			if (isCandidateForSynchronizer(aStep)){
					
				if (stepNumber(process) > 3){

					// Determine the type of the REPLY node
					String replyNodeType = getReplyNodeType(process);
					
					if (replyNodeType == null) {
						// No reply node found: don't apply the synchronizer
						SpagicLogger.error("JBI-Reply node not found and InOut component used");
						return;
					}
					
					// Create binding according to the reply node
					String synchronizerToUse = replyNodeType.equals(REPLY_STEP) ? CLASSIC_SYNCHRONIZER : HIBERNATE_TRANSACTED_SYNCHRONIZER;
					createServiceBindingForRuntime(im, synchronizerToUse);
					

					// Get The first outcoming transition to save the destination Step
					sourceTransition = getFirstOutcomingTransition(aStep);
					destinationStep = sourceTransition.getTarget();
				
					// Removing Original Transition
					removeTransitionFromProcess(process,sourceTransition);
					removeFromStepOutcomingTransition(aStep,sourceTransition);
					removeFromStepIncomingTransition(destinationStep,sourceTransition);
				
					// Creating The Synchnronizer Step
					synchronizerStep = createSynchronizerStepForStartStep(im, aStep, synchronizerToUse);
					stepToAddToProcess.add(synchronizerStep);

					// Relocate Transition
					createTransitionAndAddToProcess(process, aStep, synchronizerStep);
					createTransitionAndAddToProcess(process, synchronizerStep, destinationStep);
					createTransitionAndAddToProcess(process, synchronizerStep, aStep);
				
					syncAdded = true;
				}else{
					// No need to create a Sync relocate only the transition
					sourceTransition = getFirstOutcomingTransition(aStep);
					destinationStep = sourceTransition.getTarget();
					createTransitionAndAddToProcess(process, destinationStep, aStep);
					syncAdded = false;
				}
			}
		}// for Step Loop 
	
	
		// Change the transitions of the steps that has target reply to sync
		for (Step step : process.getSteps()) {
			List<Transition> transitionsToRemoveFromOutcomingTransition = new ArrayList<Transition>();
			for (Transition tr : step.getSourceTransitions()) {
				Step targetStep = tr.getTarget();
				if (isServiceBinding(targetStep, REPLY_STEP) ||
					isServiceBinding(targetStep, HIBERNATE_TRANSACTED_REPLY_STEP)){
					replyStepsToRemove.add(targetStep);
					if (syncAdded) {
						tr.setTarget(synchronizerStep);
						
						// Copy properties from reply node to synchronizer node, if any
						for (Map.Entry<String, Property> prop : targetStep.getProperties()) {
							BasicProperty stepProperty = stpModelfactory.createBasicProperty();
							stepProperty.setKey(prop.getKey());
							stepProperty.setValue((String)prop.getValue().getValue());
							synchronizerStep.getProperties().put(stepProperty.getKey(), stepProperty);
						}
					}
					else{
						transitionsToRemoveFromOutcomingTransition.add(tr);
					}
				}
			}//for (Transition tr :
			removeFromStepOutcomingTransition(step,transitionsToRemoveFromOutcomingTransition);
			process.getTransitions().removeAll(transitionsToRemoveFromOutcomingTransition);
		}//for (Step aStep :
		
		process.getSteps().removeAll(replyStepsToRemove);		
		process.getSteps().addAll(stepToAddToProcess);
	}
	
	public String getReplyNodeType(Process process) {
		String replyNodeType = null;
		for (Iterator<Step> it = process.getSteps().iterator(); it.hasNext() && (replyNodeType == null); ) {
			Step step = it.next();
			for (Transition tr : step.getSourceTransitions()) {
				if (isServiceBinding(tr.getTarget(), REPLY_STEP)){
					replyNodeType = REPLY_STEP;
					break;
				} else if (isServiceBinding(tr.getTarget(), HIBERNATE_TRANSACTED_REPLY_STEP)) {
					replyNodeType = HIBERNATE_TRANSACTED_REPLY_STEP;
					break;
				}
			}//for (Transition tr :
		}//for (Step aStep :
		return replyNodeType;
	}

	public String getEventTypeHandled(Step aStep){
		
		if (isServiceBinding(aStep, "FileEventBindingComponent") || isServiceBinding(aStep, "FTPEventBindingComponent")){
			return "FileEvent";
		}else if (isServiceBinding(aStep, "WebserviceEventBindingComponent")){
			return "WebserviceEvent";
		}
		return "GenericEvent";
		
	}
	public Transition findTheTransitionComingFromExternalEventHandler(Step aStep){
		
		List<Transition> incomingTransitions = aStep.getTargetTransitions();
		for (Transition incomingTransition : incomingTransitions){
			if (isIncomingTransitionFromExternalEventHandlerBindingComponent(incomingTransition))
				return incomingTransition;
		}
		return null;
	}
	
	public void setSignalEndpointURIOnStep(Process process, Step targetStep, Step signalStep){
		setPropertyOnStep(targetStep, "signalEndpointURI", SpagicUtils.getStepNameInProcess(process, signalStep));
	}
	
	public boolean isServiceBinding(Step step, String serviceBindingName) {
		return isServiceBinding(getRightStepBinding(step), serviceBindingName);
	}
	
	///
	public  void handleTimerAggregator(StpIntermediateModel im,  Process process) {
		for (Step aStep : process.getSteps()) {
			boolean isSplitter = false;
			boolean isRecipientList = false;
			// if the step is a splitter step,I search the aggregator associated to him and set a property
			if ( isServiceBinding(aStep, "JBI-XPathSplitter")) 
				isSplitter = true;
			else if (isServiceBinding(aStep,"JBI-RecipientList"))
				isRecipientList = true;
			if ( isSplitter || isRecipientList) {
				Transition transition = getFirstOutcomingTransition(aStep);
				if (transition == null)
					throw new RuntimeException(" A Splitter or RecipientList Service must have one destination ");
				int counter = 0;
				
				Step targetStep = transition.getTarget();
				List<Step> visitedSteps = new ArrayList<Step>();
				visitedSteps.add(targetStep);
				while (targetStep != null) {
					if((isSplitter && isServiceBinding(targetStep, "JBI-SplitterAggregator"))
						|| (isRecipientList && isServiceBinding(targetStep, "JBI-RecipientListAggregator"))){
						if (counter == 0){
							setPropertyOnStep(aStep, AGGREGATOR_ENDPOINT, SpagicUtils.getStepLocalNamePart(process, targetStep));
							break;
						} else 
							counter--;
					}	
					else if ((isSplitter && isServiceBinding(targetStep, "JBI-XPathSplitter"))
							|| (isRecipientList && isServiceBinding(targetStep, "JBI-RecipientList")))
						counter++;
					else if (SpagicUtils.isEndStep(targetStep))
						return;
					
					transition = targetStep.getSourceTransitions().get(0);
					targetStep = transition.getTarget();
					
					// Check that we are not in a cycle
					if (visitedSteps.contains(targetStep)) {
						// We are in a cycle
						SpagicLogger.info("Cycle encountered during preprocessing of IM in handleTimerAggregator");
						break;
					} else {
						visitedSteps.add(targetStep);
					}
				}
				
				//if exit from the cycle with the targetStep null the process hasn't been properly designed
				if (targetStep == null)
					throw new RuntimeException(" A Splitter or RecipientList must have an Aggregator associated them");
			}
		}
	}
	
}
