/**

    Copyright 2007, 2008 Engineering Ingegneria Informatica S.p.A.

    This file is part of Spagic.

    Spagic is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    any later version.

    Spagic is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
**/
package org.spagic.utils;

import java.util.HashMap;
import java.util.Map;

import javax.xml.namespace.QName;

import org.eclipse.stp.im.Process;
import org.eclipse.stp.im.Property;
import org.eclipse.stp.im.Service;
import org.eclipse.stp.im.ServiceBinding;
import org.eclipse.stp.im.Step;
import org.eclipse.stp.im.constants.ImConstants;
import org.eclipse.stp.im.runtime.ImRuntimeActivator;
import org.spagic.utils.preference.SpagicPreferenceConstants;

public class SpagicUtils {
	
	public static final String PREFIX_NAMESPACE = "urn:eng:spagic:processes:".intern();
	public static final String SPAGIC_PROPERTY_FILE = "spagic.properties".intern();
	public static final String PARAMETER_START = "${".intern();
	public static final String PARAMETER_END = "}".intern();
	public static final String VERSION_LABEL = "_v_".intern();
	
	public static final String START_SERVICE_NAME = "StartService".intern();
	public static final String END_SERVICE_NAME = "EndService".intern();

	public static String getBasePrefixNameSpace(){
		String basePrefixNameSpace = SpagicUtilsActivator.getDefault().getPluginPreferences().getString(SpagicPreferenceConstants.P_SPAGIC_BASE_NAME_SPACE);
		return basePrefixNameSpace != null ? basePrefixNameSpace : PREFIX_NAMESPACE;
	}
	
	public static String getNameSpacePrefixForProcess(Process aProcess){
		return aProcess.getName()+VERSION_LABEL+ getProcessVersion(aProcess);
	}
	public static String getNameSpaceForProcess(Process aProcess){
		return getBasePrefixNameSpace()+aProcess.getName()+":v"+getProcessVersion(aProcess);
	}
	public static QName getQNameForProcess(Process aProcess){
		String localPart = aProcess.getName()+VERSION_LABEL+getProcessVersion(aProcess);
		return new QName(getNameSpaceForProcess(aProcess), localPart);
	}
	
	public static QName getQNameForStep(Process aProcess, Step aStep){
		return  new QName(getNameSpaceForProcess(aProcess), getStepLocalNamePart(aProcess, aStep) );
	}
	
	public static String getStepLocalNamePart(Process aProcess, Step aStep){
		String localPart = aProcess.getName() + "." + aStep.getName()+VERSION_LABEL+getProcessVersion(aProcess);
		return localPart;
	}
	
	public static String getStepNameInProcess(Process aProcess, Step aStep){
		return getNameSpacePrefixForProcess(aProcess) + ":" + getStepLocalNamePart(aProcess, aStep);
	}
	
	public static String getProcessVersion(Process aProcess){
		Property prop = aProcess.getProperties().get(ImConstants.PROCESS_VERSION);
		return (prop != null ? (String)prop.getValue() : "0");
	}
	
	/**
	 * 
	 * @param s
	 * @return the right service binding for the runtime we're using during the transformation
	 */
	public static ServiceBinding getRightStepBinding(Step s, String runtimeID) {
		
		Property runtimPropertyForServiceBinding = null;
		for (ServiceBinding sb : s.getStepbindings()){
			runtimPropertyForServiceBinding = sb.getProperties().get(ImConstants.IM_STEP_BINDING_RUNTIME_ID);
			if (runtimPropertyForServiceBinding != null){
				if (runtimPropertyForServiceBinding.getValue().equals(runtimeID)){
					return sb;
				}
			}else{
				SpagicLogger.warning("ServiceBinding["+sb.getName()+"] -> for Step["+s.getName()+"] does not have a property called "+ ImConstants.IM_POOL_RUNTIME_ID);
			}
		}
		SpagicLogger.warning("Not found for Step["+s.getName()+"] a binding for runtime "+ ImConstants.IM_POOL_RUNTIME_ID);
		return null;
		
	}
	
	public static Map<String, String> getRuntimeUsedByProcess(Process aProcess) {
		HashMap<String, String> usedRuntimes = new HashMap<String, String>();
		String runtimeId = null;
		Property runtimeIdProperty = null;
		for (Step aStep : aProcess.getSteps()){
			for (ServiceBinding sb : aStep.getStepbindings() ){
				runtimeIdProperty = sb.getProperties().get(ImConstants.IM_STEP_BINDING_RUNTIME_ID);
				runtimeId = ( runtimeIdProperty != null ? (String)runtimeIdProperty.getValue() : null);
				if ((runtimeId != null) && ( usedRuntimes.get(runtimeId) == null )){
					usedRuntimes.put(runtimeId, ImRuntimeActivator.getRuntime(runtimeId).getName());
				}//--
			}//--
		}//--
		return usedRuntimes;
	}//
	
	public static boolean isProcessDeployableForRuntime(Process aProcess, String runtimeId) {
		Map<String, String> runtimesUsed = getRuntimeUsedByProcess(aProcess);
		String runtimeName = runtimesUsed.get(runtimeId);
		return (runtimeName != null);
	}//
	
	/**
	 * Return the property file name containing the dynamic parameters to use for
	 * a specific process.
	 * @param aProcess The IM process to check
	 * @return The property file name
	 */
	public static String getPropertyFileName(Process aProcess) {
		if (isDeployableConfigurable(aProcess)) {
			// The process contains some parameters: use a specific property file
			// for that process
			return getNameSpacePrefixForProcess(aProcess) + ".properties";
		} else {
			// The process doesnt contains parameters: use a fixed empty property file
			return SPAGIC_PROPERTY_FILE;
		}
	}
	
	/**
	 * Checks if a component use a dynamic configuration based on properties files.
	 * @param aProcess The IM process to check
	 * @return true if the process contains at least one dynamic parameter
	 */
	public static boolean isDeployableConfigurable(Process aProcess) {
		
		for (Step aStep : aProcess.getSteps()) {
			for (Map.Entry<String, Property> entry : aStep.getProperties()) {
				Property property = entry.getValue();
				String propertyValue = (String)property.getValue();
				if (propertyValue != null) {
					int startIndex = propertyValue.indexOf(PARAMETER_START);
					if ((startIndex != -1) && (propertyValue.indexOf(PARAMETER_END, startIndex) != -1)) {
						return true;
					}
				}
			}
		}
		
		return false;
	}
	
	public static boolean isStartStep(Step step) {
		Service serviceForStep = step.getServiceModel();
		boolean isStartStep = (serviceForStep != null ? serviceForStep.getServiceName().equalsIgnoreCase(START_SERVICE_NAME) : false);
		return isStartStep;
	}
	
	public static boolean isEndStep(Step step) {
		Service serviceForStep = step.getServiceModel();
		boolean isEndStep = (serviceForStep != null ?  serviceForStep.getServiceName().equalsIgnoreCase(END_SERVICE_NAME) : false);
		return isEndStep;
	}
	
}
